/*
 * Copyright © 2021-2023 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel24-VierConnects-core.
 *
 * ipspiel24-VierConnects-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel24-VierConnects-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel24-VierConnects-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel24.VierConnects.core.domain.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsBoard;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsDirection;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsField;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsFieldState;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsPosition;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsRow;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsRowType;

/**
 * Implements {@link VierConnectsBoard}.
 */
@SuppressWarnings("PMD.GodClass")
final class VierConnectsBoardImpl implements VierConnectsBoard {

    /**
     * The minimum number of rows.
     */
    private static final int MINIMUM_BOARD_SIZE_ROWS = 2;

    /**
     * The minimum number of columns.
     */
    private static final int MINIMUM_BOARD_SIZE_COLUMNS = 2;

    /**
     * The fields of the board, sorted first by row and then by column.
     */
    private final Map<Integer, Map<Integer, VierConnectsFieldImpl>> fields;

    /**
     * Creates a Vier Connects board.
     *
     * @param nrOfRows    The number of rows. It must be a positive number greater than or equal to 2.
     * @param nrOfColumns The number of columns. It must be a positive number greater than or equal to 2.
     * @throws IllegalArgumentException if the size does not meet the requirements.
     */
    VierConnectsBoardImpl(final int nrOfRows, final int nrOfColumns) throws IllegalArgumentException {
        if (nrOfRows < VierConnectsBoardImpl.MINIMUM_BOARD_SIZE_ROWS
                || nrOfColumns < VierConnectsBoardImpl.MINIMUM_BOARD_SIZE_COLUMNS) {
            throw new IllegalArgumentException(
                    String.format(
                            "The board row number %d is not a positive number >= %d"
                                    + " or the board column number %d is not a positive number >= %d ",
                            nrOfRows,
                            VierConnectsBoardImpl.MINIMUM_BOARD_SIZE_ROWS,
                            nrOfColumns,
                            VierConnectsBoardImpl.MINIMUM_BOARD_SIZE_COLUMNS));
        }

        this.fields = new LinkedHashMap<>();
        for (int rowIndex = 0; rowIndex < nrOfRows; ++rowIndex) {
            final Map<Integer, VierConnectsFieldImpl> row = new HashMap<>();
            this.fields.put(rowIndex, row);

            for (int columnIndex = 0; columnIndex < nrOfColumns; ++columnIndex) {
                row.put(
                        columnIndex,
                        new VierConnectsFieldImpl(
                                this,
                                VierConnectsPosition.of(rowIndex, columnIndex),
                                VierConnectsFieldState.EMPTY));
            }
        }
    }

    /**
     * Copies an VierConnects board.
     *
     * @param source The board to copy.
     */
    private VierConnectsBoardImpl(final VierConnectsBoardImpl source) {
        Objects.requireNonNull(source, "source");

        final int nrOfRows = source.getRowSize();
        final int nrOfColumns = source.getColumnSize();
        this.fields = new LinkedHashMap<>();
        for (int rowIndex = 0; rowIndex < nrOfRows; ++rowIndex) {
            final Map<Integer, VierConnectsFieldImpl> row = new HashMap<>();
            this.fields.put(rowIndex, row);

            for (int columnIndex = 0; columnIndex < nrOfColumns; ++columnIndex) {
                row.put(
                        columnIndex,
                        new VierConnectsFieldImpl(
                                this,
                                VierConnectsPosition.of(rowIndex, columnIndex),
                                source.getFieldAtInternal(VierConnectsPosition.of(rowIndex, columnIndex)).orElseThrow()
                                        .getState()));
            }
        }
    }

    @Override
    public String toString() {
        final StringBuilder result = new StringBuilder();
        for (int row = 0; row < this.fields.size(); ++row) {
            final Map<Integer, VierConnectsFieldImpl> rowMap = this.fields.get(row);
            for (int column = 0; column < rowMap.size(); ++column) {
                result.append(rowMap.get(column).getState().toString()).append(' ');
            }
            result.append('\n');
        }
        return result.toString();
    }

    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof VierConnectsBoardImpl) {
            final VierConnectsBoardImpl other = (VierConnectsBoardImpl) obj;
            return this.fields.equals(other.fields);
        }
        return false;
    }

    @Override
    public int hashCode() {
        return this.fields.hashCode();
    }

    @Override
    public int getRowSize() {
        return this.fields.size();
    }

    @Override
    public int getColumnSize() {
        // Überprüft, ob das Feld leer ist, um eine NullPointer-Exception zu vermeiden.
        if (this.fields.isEmpty()) {
            return 0;
        }
        // Nimmt an, dass alle Zeilen dieselbe Anzahl von Spalten haben und gibt die Größe der ersten Zeile zurück.
        final Map.Entry<Integer,
                Map<Integer, VierConnectsFieldImpl>> firstEntry = this.fields.entrySet().iterator().next();
        return firstEntry.getValue().size();

//        return this.fields.size();
    }

    @Override
    public boolean hasFieldAt(final VierConnectsPosition position) {
        final int nrOfRows = this.getRowSize();
        final int nrOfColumns = this.getColumnSize();
        final int row = position.getRow();
        final int column = position.getColumn();
        return row >= 0 && row < nrOfRows && column >= 0 && column < nrOfColumns;
    }

    @Override
    public VierConnectsFieldImpl getFieldAt(final VierConnectsPosition position) {
        if (!this.hasFieldAt(position)) {
            throw new IllegalArgumentException(String.format("Position %s out of range.", position));
        }

        return this.getFieldAtInternal(position).orElseThrow();
    }

    @Override
    public List<List<? extends VierConnectsField>> getFields() {
        final List<List<? extends VierConnectsField>> result = new ArrayList<>();
        this.fields.values()
                .forEach((final Map<Integer, VierConnectsFieldImpl> row) -> result.add(new ArrayList<>(row.values())));
        return result;
    }

    @Override
    public Set<VierConnectsRow> getRowsWithFourMarked() {
        final Set<VierConnectsRow> result = new LinkedHashSet<>();
        for (int rowIndex = 0; rowIndex < this.getRowSize(); ++rowIndex) {
            addRowIfFourMarked(result, this.getRowOfFields(VierConnectsRowType.ROW, rowIndex));
        }
        for (int columnIndex = 0; columnIndex < this.getColumnSize(); ++columnIndex) {
            addRowIfFourMarked(result, this.getRowOfFields(VierConnectsRowType.COLUMN, columnIndex));
        }
        for (int diagonalIndexRow = 0; diagonalIndexRow < this.getRowSize(); ++diagonalIndexRow) {
            addRowIfFourMarked(result, this.getRowOfFields(VierConnectsRowType.DIAGONAL, diagonalIndexRow, true));
        }
        for (int diagonalIndexColumn = 0; diagonalIndexColumn < this.getColumnSize(); ++diagonalIndexColumn) {
            addRowIfFourMarked(result, this.getRowOfFields(VierConnectsRowType.DIAGONAL, diagonalIndexColumn, false));
        }
        for (int diagonalIndexRow = this.getRowSize() - 1; diagonalIndexRow >= 0; --diagonalIndexRow) {
            addRowIfFourMarked(result, this.getRowOfFields(VierConnectsRowType.DIAGONALALT, diagonalIndexRow, true));
        }
        for (int diagonalIndexColumn = this.getColumnSize() - 1; diagonalIndexColumn >= 0; --diagonalIndexColumn) {
            addRowIfFourMarked(result, this.getRowOfFields(VierConnectsRowType.DIAGONALALT,
                    diagonalIndexColumn, false));
        }
        return result;
    }

    @Override
    public Map<VierConnectsPosition, VierConnectsFieldImpl> getFieldsBeing(final VierConnectsFieldState fieldState) {
        final Map<VierConnectsPosition, VierConnectsFieldImpl> result = new LinkedHashMap<>();
        final int nrOfRows = this.getRowSize();
        final int nrOfColumns = this.getColumnSize();
        for (int rowIndex = 0; rowIndex < nrOfRows; ++rowIndex) {
            for (int columnIndex = 0; columnIndex < nrOfColumns; ++columnIndex) {
                final VierConnectsPosition position = VierConnectsPosition.of(rowIndex, columnIndex);
                final VierConnectsFieldImpl field = this.getFieldAtInternal(position).orElseThrow();
                if (field.getState().equals(fieldState)) {
                    result.put(position, field);
                }
            }
        }

        return result;
    }

    /**
     * Adds a row to a set of rows if it has 4 of the same type in a line.
     *
     * @param result The result set.
     * @param row    The row to add.
     */
    private static void addRowIfFourMarked(final Set<VierConnectsRow> result, final VierConnectsRow row) {
        final Optional<VierConnectsFieldState> rowState = row.getState();
        if (rowState.isPresent() && !rowState.get().equals(VierConnectsFieldState.EMPTY)) {
            result.add(row);
        }
    }

    @Override
    public VierConnectsBoardImpl deepCopy() {
        return new VierConnectsBoardImpl(this);
    }

    /**
     * Returns a row of fields given the row type and the row index.
     *
     * @param type  The row type.
     * @param index The row index.
     */
    private VierConnectsRow getRowOfFields(final VierConnectsRowType type, final int index) {
        final List<? extends VierConnectsField> fieldsOfRow;

        switch (type) {
        case ROW:
            fieldsOfRow = this.getFields(VierConnectsPosition.of(index, 0), VierConnectsDirection.EAST);
            break;
        case COLUMN:
            fieldsOfRow = this.getFields(VierConnectsPosition.of(0, index), VierConnectsDirection.SOUTH);
            break;
        default:
            throw new IllegalArgumentException(String.format("Unknown VierConnectsRowType %s.", type));
        }

        return new VierConnectsRowImpl(type, index, fieldsOfRow);
    }
    
    /**
     * Returns a row of fields given the row type and the row index.
     *
     * @param type  The row type.
     * @param index The row index.
     * @param isBelowDisecting If below disecting.
     */
    private VierConnectsRow getRowOfFields(final VierConnectsRowType type,
            final int index,
            final boolean isBelowDisecting) {
        final List<? extends VierConnectsField> fieldsOfRow;

        switch (type) {
        case DIAGONAL:
            if (isBelowDisecting) {
                fieldsOfRow = this.getFields(VierConnectsPosition.of(index, 0), VierConnectsDirection.SOUTHEAST);
            } else {
                fieldsOfRow = this.getFields(VierConnectsPosition.of(0, index), VierConnectsDirection.SOUTHEAST);
            }
            break;
        case DIAGONALALT:
            if (isBelowDisecting) {
                fieldsOfRow = this.getFields(VierConnectsPosition.of(index, this.getColumnSize() - 1),
                        VierConnectsDirection.SOUTHWEST);
            } else {
                fieldsOfRow = this.getFields(VierConnectsPosition.of(0, index),
                        VierConnectsDirection.SOUTHWEST);
            }
            break;
        default:
            throw new IllegalArgumentException(String.format("Unknown VierConnectsRowType %s.", type));
        }

        return new VierConnectsRowImpl(type, index, fieldsOfRow);
    }

//    /**
//     * Returns a row of fields that is diagonal, the index will count along
//     * the rows if isBelowDisecting is true, along the columns if it is false.
//     *
//     * @param isBelowDisecting If the diagonals are counted along the columns or the rows.
//     * @param index            The row index.
//     */
//    private VierConnectsRow getRowOfFieldsDiagonal(final boolean isBelowDisecting, final int index) {
//        final List<? extends VierConnectsField> fieldsOfRow;
//        if (isBelowDisecting) {
//            fieldsOfRow = this.getFields(VierConnectsPosition.of(index, 0), VierConnectsDirection.SOUTHEAST);
//        } else {
//            fieldsOfRow = this.getFields(VierConnectsPosition.of(0, index), VierConnectsDirection.SOUTHEAST);
//        }
//
//        return new VierConnectsRowImpl(VierConnectsRowType.DIAGONAL, index, fieldsOfRow);
//    }
//
//    /**
//     * Returns a row of fields that is diagonal, the index will count along
//     * the rows if isBelowDisecting is true, along the columns if it is false.
//     *
//     * @param isBelowDisecting If the diagonals are counted along the columns or the rows.
//     * @param index            The row index.
//     */
//    private VierConnectsRow getRowOfFieldsDiagonalAlt(final boolean isBelowDisecting, final int index) {
//        final List<? extends VierConnectsField> fieldsOfRow;
//        if (isBelowDisecting) {
//            fieldsOfRow = this.getFields(VierConnectsPosition.of(index, this.getColumnSize() - 1),
//                    VierConnectsDirection.SOUTHWEST);
//        } else {
//            fieldsOfRow = this.getFields(VierConnectsPosition.of(0, index),
//                    VierConnectsDirection.SOUTHWEST);
//        }
//
//        return new VierConnectsRowImpl(VierConnectsRowType.DIAGONAL, index, fieldsOfRow);
//    }

    /**
     * Returns a list of fields starting at a given position and extending towards a given direction till the board's
     * end.
     *
     * @param start     The starting position.
     * @param direction The direction.
     */
    private List<? extends VierConnectsField> getFields(final VierConnectsPosition start,
            final VierConnectsDirection direction) {
        final List<VierConnectsField> result = new ArrayList<>();

        VierConnectsPosition pos = start;
        Optional<VierConnectsFieldImpl> field = this.getFieldAtInternal(pos);
        while (field.isPresent()) {
            result.add(field.get());
            pos = direction.step(pos);
            field = this.getFieldAtInternal(pos);
        }

        return result;
    }

    /**
     * Returns a field at a given position. If the position does not exist, {@link Optional#empty()} is returned.
     *
     * @param position The position of the field.
     * @return The field.
     */
    private Optional<VierConnectsFieldImpl> getFieldAtInternal(final VierConnectsPosition position) {
        return Optional.ofNullable(this.fields.get(position.getRow()))
                .map((final Map<Integer, VierConnectsFieldImpl> map) -> map.get(position.getColumn()));
    }
}
