/*
 * Copyright © 2021-2023 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel24-VierConnects-core.
 *
 * ipspiel24-VierConnects-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel24-VierConnects-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel24-VierConnects-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel24.VierConnects.core.domain.impl;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.core.domain.PlayerState;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsBoard;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsField;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsFieldState;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsPlayer;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsPosition;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsRow;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsState;

/**
 * Implements {@link VierConnectsState}.
 */
@SuppressWarnings("PMD.GodClass")
final class VierConnectsStateImpl implements VierConnectsState {

    /**
     * The board.
     */
    private final VierConnectsBoard board;

    /**
     * The player using crosses.
     */
    private final VierConnectsPlayer crossesPlayer;

    /**
     * The player using noughts.
     */
    private final VierConnectsPlayer noughtsPlayer;

    /**
     * The current player.
     */
    private VierConnectsPlayer currentPlayer;

    /**
     * Creates an Vier Connects state.
     *
     * @param board         The board.
     * @param crossesPlayer The player using crosses.
     * @param noughtsPlayer The player using noughts.
     * @param crossesIsNext {@code true} if the player using crosses is next, else the player using noughts is next.
     * @throws GameException if the state cannot be created according to the rules of the game.
     */
    VierConnectsStateImpl(final VierConnectsBoard board, final VierConnectsPlayer crossesPlayer,
            final VierConnectsPlayer noughtsPlayer, final boolean crossesIsNext) throws GameException {

        this.board = Objects.requireNonNull(board, "board");
        this.crossesPlayer = Objects.requireNonNull(crossesPlayer, "crossesPlayer");
        this.noughtsPlayer = Objects.requireNonNull(noughtsPlayer, "noughtsPlayer");
        this.currentPlayer = crossesIsNext ? this.crossesPlayer : this.noughtsPlayer;

        if (!this.crossesPlayer.isUsingCrosses()) {
            throw new IllegalArgumentException(String.format("Player %s does not use crosses.", this.crossesPlayer));
        }
        if (this.noughtsPlayer.isUsingCrosses()) {
            throw new IllegalArgumentException(String.format("Player %s does not use noughts.", this.noughtsPlayer));
        }
        if (this.crossesPlayer.getName().equals(this.noughtsPlayer.getName())) {
            throw new IllegalArgumentException(
                    String.format("Both players have the same name '%s'.", this.crossesPlayer.getName()));
        }
    }

    /**
     * Creates an VierConnects state by copying an existing one.
     *
     * @param source The state to copy.
     */
    VierConnectsStateImpl(final VierConnectsStateImpl source) {
        this.board = source.board.deepCopy();
        this.crossesPlayer = source.crossesPlayer.deepCopy();
        this.noughtsPlayer = source.noughtsPlayer.deepCopy();
        this.currentPlayer = source.isCrossesPlayerCurrent() ? this.crossesPlayer : this.noughtsPlayer;
    }

    @Override
    public String toString() {
        return String.format(
                "VierConnectsState[board=\n%s, crossesPlayer=%s, noughtsPlayer=%s, currentPlayer=%s]",
                this.board,
                this.crossesPlayer,
                this.noughtsPlayer,
                this.currentPlayer.isUsingCrosses() ? VierConnectsFieldState.CROSS : VierConnectsFieldState.NOUGHT);
    }

    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof VierConnectsStateImpl) {
            final VierConnectsStateImpl other = (VierConnectsStateImpl) obj;
            return this.board.equals(other.board) && this.crossesPlayer.equals(other.crossesPlayer)
                    && this.noughtsPlayer.equals(other.noughtsPlayer)
                    && this.isCrossesPlayerCurrent() == other.isCrossesPlayerCurrent();
        }
        return false;
    }

    @Override
    public int hashCode() {
        return Objects.hash(this.board, this.crossesPlayer, this.noughtsPlayer, this.isCrossesPlayerCurrent());
    }

    @Override
    public VierConnectsStateImpl deepCopy() {
        return new VierConnectsStateImpl(this);
    }

    @Override
    public Map<String, VierConnectsPlayer> getPlayers() {
        final Map<String, VierConnectsPlayer> result = new LinkedHashMap<>();
        result.put(this.crossesPlayer.getName(), this.crossesPlayer);
        result.put(this.noughtsPlayer.getName(), this.noughtsPlayer);
        return result;
    }

    @Override
    public VierConnectsBoard getBoard() {
        return this.board;
    }

    @Override
    public VierConnectsPlayer getCrossesPlayer() {
        return this.crossesPlayer;
    }

    @Override
    public VierConnectsPlayer getNoughtsPlayer() {
        return this.noughtsPlayer;
    }

    @Override
    public VierConnectsPlayer getCurrentPlayer() {
        return this.currentPlayer;
    }

    
    @Override
    public void moveCompleted() {
        final Set<VierConnectsRow> winningRows = this.board.getRowsWithFourMarked();
        if (!winningRows.isEmpty()) {
            final VierConnectsFieldState rowState = winningRows.iterator().next().getState().orElseThrow();
            this.gameOver(rowState);
        } else {
            final Map<VierConnectsPosition,
                    ? extends VierConnectsField> emptyFields = this.board.getFieldsBeing(VierConnectsFieldState.EMPTY);
            if (emptyFields.isEmpty()) {
                this.gameOver(VierConnectsFieldState.EMPTY);
            }
        }
    }

    @Override
    public Set<VierConnectsPlayer> computeNextPlayers() {
        if (this.currentPlayer.getState().equals(PlayerState.PLAYING)) {
            return Collections.singleton(this.currentPlayer);
        }
        return Collections.emptySet();
    }

    @Override
    public void nextTurn() {
        this.currentPlayer = this.getOtherPlayer();
    }

    /**
     * Returns the currently inactive player.
     */
    private VierConnectsPlayer getOtherPlayer() {
        if (this.isCrossesPlayerCurrent()) {
            return this.noughtsPlayer;
        }
        return this.crossesPlayer;
    }

    /**
     * Returns {@code true} iff the current player uses crosses.
     */
    @SuppressWarnings("PMD.CompareObjectsWithEquals")
    private boolean isCrossesPlayerCurrent() {
        return this.currentPlayer == this.crossesPlayer;
    }

    /**
     * Updates the player states after the game has finished.
     *
     * @param rowState The state of some uniformly marked row, or {@link VierConnectsFieldState#EMPTY} if no player has
     *                 won
     *                 the game yet and no empty fields have remained.
     */
    private void gameOver(final VierConnectsFieldState rowState) {
        switch (rowState) {
        case CROSS:
            this.crossesPlayer.setState(PlayerState.WON);
            this.noughtsPlayer.setState(PlayerState.LOST);
            break;
        case NOUGHT:
            this.crossesPlayer.setState(PlayerState.LOST);
            this.noughtsPlayer.setState(PlayerState.WON);
            break;
        case EMPTY:
            this.crossesPlayer.setState(PlayerState.DRAW);
            this.noughtsPlayer.setState(PlayerState.DRAW);
            break;
        default:
            throw new IllegalArgumentException(
                    String.format("VierConnectsState.gameOver() called with state %s.", rowState));
        }
    }
}
