/*
 * Copyright © 2021-2023 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel24-VierConnects-core.
 *
 * ipspiel24-VierConnects-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel24-VierConnects-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel24-VierConnects-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel24.VierConnects.core.moves.impl;

import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsDirection;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsField;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsFieldState;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsPlayer;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsPosition;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsState;

/**
 * Represents a move which places a mark on a field on the board.
 * <p>
 * Note that this move is only allowed if the field is empty.
 */
final class VierConnectsPlaceMarkMove extends AbstractVierConnectsMove {

    /**
     * {@code true} if a cross is to be made, {@code false} chooses a nought instead.
     */
    private final boolean cross;
    /**
     * The position of where the player wants to place the token.
     */
    private final int columnNr;


    /**
     * Creates an {@link VierConnectsPlaceMarkMove} object.
     *
     * @param cross         {@code true} if a cross is to be made, {@code false} chooses a nought instead.
     * @param column The column of the field having its
     *  state changed. It must point to an empty field.
     */
    VierConnectsPlaceMarkMove(final boolean cross, final int column) {
        this.cross = cross;
        this.columnNr = column;
        
    }

    /**
     * Returns {@code true} if a cross is to be made, and {@code false} if a nought is to be made instead.
     */
    boolean isCross() {
        return this.cross;
    }

    /**
     * Returns the position of the token placed on the board.
     */
    int getColoumnNr() {
        return this.columnNr;
    }
    
    /**
     * Returns the next token position.
     *     @param state
     */
    VierConnectsPosition getNewTokenPosition(final VierConnectsState state) throws GameException {
        final VierConnectsPosition tokenPosition = VierConnectsPosition.of(0, this.getColoumnNr());

        VierConnectsField fieldToPlaceOn = state.getBoard().getFieldAt(tokenPosition);
        
        if (!fieldToPlaceOn.getState().equals(VierConnectsFieldState.EMPTY)) {
            throw new GameException(
                    String.format(
                            "Player cannot make %s on field at %s as the column is already full.",
                            toFieldState(this.cross),
                            this.getColoumnNr()));
        }
        
        VierConnectsField nextPossibleField;
        final int nrOfRows = state.getBoard().getRowSize();
        for (int rowIndex = 0; rowIndex < nrOfRows; rowIndex++) {
            
            if (fieldToPlaceOn.hasNeighbour(VierConnectsDirection.SOUTH)) {
                nextPossibleField = fieldToPlaceOn.getNeighbour(VierConnectsDirection.SOUTH);
            } else {
                return fieldToPlaceOn.getPosition();
            }
            
            if (!nextPossibleField.getState().equals(VierConnectsFieldState.EMPTY)) {
                return fieldToPlaceOn.getPosition();
            } else {
                fieldToPlaceOn = nextPossibleField;
            }
        }
        
        
        throw new GameException("You should not be here. The placing has gone very wrong.");
    }
    

    @Override
    public void applyTo(final VierConnectsState state, final VierConnectsPlayer player) throws GameException {
        if (this.cross != player.isUsingCrosses()) {
            throw new GameException(
                    String.format(
                            "Player %s cannot make %s on field at %s.",
                            player,
                            toFieldState(this.cross),
                            this.getColoumnNr()));
        }
        
        final VierConnectsPosition realNewPosition = getNewTokenPosition(state); 

        state.getBoard().getFieldAt(realNewPosition).changeState(toFieldState(this.cross));
        state.moveCompleted();
    }

    @Override
    public String toString() {
        return String.format("%s on field at %s", toFieldState(this.cross), this.columnNr);
    }

    /**
     * Maps a cross/nought boolean to a target {@link VierConnectsFieldState}.
     *
     * @param cross {@code true} for a cross and {@code false} for a nought.
     */
    private static VierConnectsFieldState toFieldState(final boolean cross) {
        return cross ? VierConnectsFieldState.CROSS : VierConnectsFieldState.NOUGHT;
    }
}
